import numpy as np
import scipy.special as special
import spharpy.special as _special


def acn2nm(acn):
    r"""
    Calculate the spherical harmonic order n and degree m for a linear
    coefficient index, according to the Ambisonics Channel Convention [1]_.

    .. math::

        n = \lfloor \sqrt{acn + 1} \rfloor - 1

        m = acn - n^2 -n


    References
    ----------
    .. [1]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1–11, 2011.


    Parameters
    ----------
    n : integer, ndarray
        Spherical harmonic order
    m : integer, ndarray
        Spherical harmonic degree

    Returns
    -------
    acn : integer, ndarray
        Linear index

    """
    acn = np.asarray(acn, dtype=int)

    n = (np.ceil(np.sqrt(acn + 1)) - 1)
    m = acn - n**2 - n

    n = n.astype(int, copy=False)
    m = m.astype(int, copy=False)

    return n, m


def nm2acn(n, m):
    """
    Calculate the linear index coefficient for a spherical harmonic order n
    and degree m, according to the Ambisonics Channel Convention [1]_.

    .. math::

        acn = n^2 + n + m

    References
    ----------
    .. [1]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1–11, 2011.


    Parameters
    ----------
    n : integer, ndarray
        Spherical harmonic order
    m : integer, ndarray
        Spherical harmonic degree

    Returns
    -------
    acn : integer, ndarray
        Linear index

    """
    n = np.asarray(n, dtype=int)
    m = np.asarray(m, dtype=int)

    if n.size != m.size:
        raise ValueError("n and m need to be of the same size")

    return n**2 + n + m


def spherical_harmonic_basis(n_max, coords):
    r"""
    Calulcates the complex valued spherical harmonic basis matrix of order Nmax
    for a set of points given by their elevation and azimuth angles.
    The spherical harmonic functions are fully normalized (N3D) and include the
    Condon-Shotley phase term :math:`(-1)^m` [2]_, [3]_.

    .. math::

        Y_n^m(\theta, \phi) = \sqrt{\frac{2n+1}{4\pi}
        \frac{(n-m)!}{(n+m)!}} P_n^m(\cos \theta) e^{i m \phi}

    References
    ----------
    .. [2]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [3]  B. Rafaely, Fundamentals of Spherical Array Processing, vol. 8.
            Springer, 2015.


    Parameters
    ----------
    n_max : integer
        Spherical harmonic order
    coordinates : Coordinates
        Coordinate object with sampling points for which the basis matrix is
        calculated

    Returns
    -------
    Y : double, ndarray, matrix
        Complex spherical harmonic basis matrix
    """

    n_coeff = (n_max+1)**2

    basis = np.zeros((coords.n_points, n_coeff), dtype=complex)

    for acn in range(n_coeff):
        order, degree = acn2nm(acn)
        basis[:, acn] = _special.spherical_harmonic(
            order,
            degree,
            coords.elevation,
            coords.azimuth)

    return basis


def spherical_harmonic_basis_gradient(n_max, coords):
    r"""
    Calulcates the gradient on the unit sphere of the complex valued spherical
    harmonic basis matrix of order N for a set of points given by their
    elevation and azimuth angles.
    The spherical harmonic functions are fully normalized (N3D) and include the
    Condon-Shotley phase term :math:`(-1)^m` [2]_. This implementation avoids
    singularities at the poles using identities derived in [5]_.


    References
    ----------
    .. [2]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [9]  J. Du, C. Chen, V. Lesur, and L. Wang, “Non-singular spherical
            harmonic expressions of geomagnetic vector and gradient tensor
            fields in the local north-oriented reference frame,” Geoscientific
            Model Development, vol. 8, no. 7, pp. 1979–1990, Jul. 2015.


    Parameters
    ----------
    n_max : integer
        Spherical harmonic order
    coordinates : Coordinates
        Coordinate object with sampling points for which the basis matrix is
        calculated

    Returns
    -------
    grad_elevation : double, ndarray, matrix
        Gradient with regard to the elevation angle.
    grad_azimuth : double, ndarray, matrix
        Gradient with regard to the azimuth angle.


    """
    n_points = coords.n_points
    n_coeff = (n_max+1)**2
    theta = coords.elevation
    phi = coords.azimuth
    grad_theta = np.zeros((n_points, n_coeff), dtype=complex)
    grad_phi = np.zeros((n_points, n_coeff), dtype=complex)

    for acn in range(n_coeff):
        n, m = acn2nm(acn)

        grad_theta[:, acn] = \
            _special.spherical_harmonic_derivative_theta(
                n, m, theta, phi)
        grad_phi[:, acn] = \
            _special.spherical_harmonic_gradient_phi(
                n, m, theta, phi)

    return grad_theta, grad_phi


def spherical_harmonic_basis_real(n_max, coords):
    r"""
    Calulcates the real valued spherical harmonic basis matrix of order Nmax
    for a set of points given by their elevation and azimuth angles.
    The spherical harmonic functions are fully normalized (N3D) and follow
    the AmbiX phase convention [1]_.

    .. math::

        Y_n^m(\theta, \phi) = \sqrt{\frac{2n+1}{4\pi}
        \frac{(n-|m|)!}{(n+|m|)!}} P_n^{|m|}(\cos \theta)
        \begin{cases}
            \displaystyle \cos(|m|\phi),  & \text{if $m \ge 0$} \newline
            \displaystyle \sin(|m|\phi) ,  & \text{if $m < 0$}
        \end{cases}

    References
    ----------
    .. [1]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1–11, 2011.


    Parameters
    ----------
    n : integer
        Spherical harmonic order
    coordinates : Coordinates
        Coordinate object with sampling points for which the basis matrix is
        calculated

    Returns
    -------
    Y : double, ndarray, matrix
        Real valued spherical harmonic basis matrix


    """
    n_coeff = (n_max+1)**2

    basis = np.zeros((coords.n_points, n_coeff), dtype=float)

    for acn in range(n_coeff):
        order, degree = acn2nm(acn)
        basis[:, acn] = _special.spherical_harmonic_real(
            order,
            degree,
            coords.elevation,
            coords.azimuth)

    return basis


def spherical_harmonic_basis_gradient_real(n_max, coords):
    r"""
    Calulcates the gradient on the unit sphere of the real valued spherical
    harmonic basis matrix of order N for a set of points given by their
    elevation and azimuth angles.
    The spherical harmonic functions are fully normalized (N3D) and follow
    the AmbiX phase convention [1]_. This implementation avoids
    singularities at the poles using identities derived in [5]_.


    References
    ----------
    .. [1]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1–11, 2011.
    .. [9]  J. Du, C. Chen, V. Lesur, and L. Wang, “Non-singular spherical
            harmonic expressions of geomagnetic vector and gradient tensor
            fields in the local north-oriented reference frame,” Geoscientific
            Model Development, vol. 8, no. 7, pp. 1979–1990, Jul. 2015.

    Parameters
    ----------
    n_max : integer
        Spherical harmonic order
    coordinates : Coordinates
        Coordinate object with sampling points for which the basis matrix is
        calculated

    Returns
    -------
    Y : double, ndarray, matrix
        Complex spherical harmonic basis matrix

    """
    n_points = coords.n_points
    n_coeff = (n_max+1)**2
    theta = coords.elevation
    phi = coords.azimuth
    grad_theta = np.zeros((n_points, n_coeff), dtype=float)
    grad_phi = np.zeros((n_points, n_coeff), dtype=float)

    for acn in range(n_coeff):
        n, m = acn2nm(acn)

        grad_theta[:, acn] = \
            _special.spherical_harmonic_derivative_theta_real(
                n, m, theta, phi)
        grad_phi[:, acn] = \
            _special.spherical_harmonic_gradient_phi_real(
                n, m, theta, phi)

    return grad_theta, grad_phi


def modal_strength(n_max,
                   kr,
                   arraytype='rigid'):
    r"""
    Modal strenght function for microphone arrays.

    .. math::

        b(kr) =
        \begin{cases}
            \displaystyle 4\pi i^n j_n(kr),  & \text{open} \newline
            \displaystyle  4\pi i^{(n-1)} \frac{1}{(kr)^2 h_n^\prime(kr)},
                & \text{rigid} \newline
            \displaystyle  4\pi i^n (j_n(kr) - i j_n^\prime(kr)),
                & \text{cardioid}
        \end{cases}


    Notes
    -----
    This implementation uses the second order Hankel function, see [4]_ for an
    overview of the corresponding sign conventions.

    References
    ----------
    .. [4]  V. Tourbabin and B. Rafaely, “On the Consistent Use of Space and
            Time Conventions in Array Processing,” vol. 101, pp. 470–473, 2015.


    Parameters
    ----------
    n : integer, ndarray
        Spherical harmonic order
    kr : double, ndarray
        Wave number * radius
    arraytype : string
        Array configuration. Can be a microphones mounted on a rigid sphere,
        on a virtual open sphere or cardioid microphones on an open sphere.

    Returns
    -------
    B : double, ndarray
        Modal strength diagonal matrix

    """
    n_coeff = (n_max+1)**2
    n_bins = kr.shape[0]

    modal_strength_mat = np.zeros((n_bins, n_coeff, n_coeff), dtype=complex)

    for n in range(n_max+1):
        bn = _modal_strength(n, kr, arraytype)
        for m in range(-n, n+1):
            acn = n*n + n + m
            modal_strength_mat[:, acn, acn] = bn

    return np.squeeze(modal_strength_mat)


def _modal_strength(n, kr, config):
    """Helper function for the calculation of the modal strength for
    plane waves"""
    if config == 'open':
        ms = 4*np.pi*pow(1.0j, n) * _special.spherical_bessel(n, kr)
    elif config == 'rigid':
        ms = 4*np.pi*pow(1.0j, n+1) / \
            _special.spherical_hankel(n, kr, derivative=True) / (kr)**2
    elif config == 'cardioid':
        ms = 4*np.pi*pow(1.0j, n) * \
            (_special.spherical_bessel(n, kr) -
                1.0j * _special.spherical_bessel(n, kr, derivative=True))
    else:
        raise ValueError("Invalid configuration.")

    return ms


def aperture_vibrating_spherical_cap(
        n_max,
        rad_sphere,
        rad_cap):
    r"""
    Aperture function for a vibrating cap with radius :math:`r_c` in a rigid
    sphere with radius :math:`r_s` [5]_, [6]_

    .. math::

        a_n (r_{s}, \alpha) = 4 \pi
        \begin{cases}
            \displaystyle \left(2n+1\right)\left[
                P_{n-1} \left(\cos\alpha\right) -
                P_{n+1} \left(\cos\alpha\right) \right],
                & {n>0} \newline
            \displaystyle  (1 - \cos\alpha)/2,  & {n=0}
        \end{cases}

    where :math:`\alpha = \arcsin \left(\frac{r_c}{r_s} \right)` is the
    aperture angle.

    Parameters
    ----------
    n_max : integer, ndarray
        Maximal spherical harmonic order
    r_sphere : double, ndarray
        Radius of the sphere
    r_cap : double
        Radius of the vibrating cap

    Returns
    -------
    A : double, ndarray
        Aperture function in diagonal matrix form with shape
        :math:`[(n_{max}+1)^2~\times~(n_{max}+1)^2]`

    References
    ----------
    .. [5]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [6]  B. Rafaely and D. Khaykin, “Optimal Model-Based Beamforming and
            Independent Steering for Spherical Loudspeaker Arrays,” IEEE
            Transactions on Audio, Speech, and Language Processing, vol. 19,
            no. 7, pp. 2234-2238, 2011

    Notes
    -----
    Eq. (3) in Ref. [6]_ contains an error, here, the power of 2 on pi is
    omitted on the normalization term.

    """
    angle_cap = np.arcsin(rad_cap / rad_sphere)
    arg = np.cos(angle_cap)
    n_sh = (n_max+1)**2

    aperture = np.zeros((n_sh, n_sh), dtype=float)

    aperture[0, 0] = (1-arg)*2*np.pi
    for n in range(1, n_max+1):
        legendre_minus = special.legendre(n-1)(arg)
        legendre_plus = special.legendre(n+1)(arg)
        legendre_term = legendre_minus - legendre_plus
        for m in range(-n, n+1):
            acn = nm2acn(n, m)
            aperture[acn, acn] = legendre_term * 4 * np.pi / (2*n+1)

    return aperture


def radiation_from_sphere(
        n_max,
        rad_sphere,
        k,
        distance,
        density_medium=1.2,
        speed_of_sound=343.0):
    r"""
    Radiation function in SH for a vibrating sphere including the radiation
    impedance and the propagation to a arbitrary distance from the sphere.
    The sign and phase conventions result in a positive pressure response for
    a positive cap velocity with the intensity vector pointing away from the
    source.

    TODO: This function does not have a test yet.

    References
    ----------
    .. [7]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [8]  F. Zotter, A. Sontacchi, and R. Höldrich, “Modeling a spherical
            loudspeaker system as multipole source,” in Proceedings of the 33rd
            DAGA German Annual Conference on Acoustics, 2007, pp. 221-222.


    Parameters
    ----------
    n_max : integer, ndarray
        Maximal spherical harmonic order
    r_sphere : double, ndarray
        Radius of the sphere
    k : double, ndarray
        Wave number
    distance : double
        Distance from the origin
    density_medium : double
        Density of the medium surrounding the sphere. Default is 1.2 for air.
    speed_of_sound : double
        Speed of sound in m/s

    Returns
    -------
    R : double, ndarray
        Radiation function in diagonal matrix form with shape
        :math:`[K \times (n_{max}+1)^2~\times~(n_{max}+1)^2]`

    """
    n_sh = (n_max+1)**2

    k = np.atleast_1d(k)
    n_bins = k.shape[0]
    radiation = np.zeros((n_bins, n_sh, n_sh), dtype=complex)

    for n in range(n_max+1):
        hankel = _special.spherical_hankel(n, k*distance, kind=2)
        hankel_prime = _special.spherical_hankel(
            n, k*rad_sphere, kind=2, derivative=True)
        radiation_order = -1j * hankel/hankel_prime * \
            density_medium * speed_of_sound
        for m in range(-n, n+1):
            acn = nm2acn(n, m)
            radiation[:, acn, acn] = radiation_order

    return radiation



"""
Tests for modal strength function
"""
from filehandling import read_matrix_from_mat
import spharpy.spherical as sh
import numpy as np


def test_modal_strength_open():
    n_max = 5
    n_bins = 128
    k = np.linspace(0.5, 15, n_bins)
    reference = read_matrix_from_mat('./tests/data/modal_strength_open.mat')
    bn = sh.modal_strength(n_max, k, arraytype='open')
    np.testing.assert_allclose(bn, reference)


def test_modal_strength_rigid():
    n_max = 5
    n_bins = 128
    k = np.linspace(0.5, 15, n_bins)
    # ref data has wrong sign (calculated as i^(n-1) instead of i^(n+1)
    reference = -read_matrix_from_mat('./tests/data/modal_strength_rigid.mat')
    bn = sh.modal_strength(n_max, k, arraytype='rigid')
    np.testing.assert_allclose(bn, reference)



from .spatial import (
    scatter,
    pcolor_map,
    pcolor_sphere,
    balloon,
    balloon_wireframe,
    voronoi_cells_sphere,
    contour,
    contour_map,
    MidpointNormalize,
)

from .cmap import phase_twilight


__all__ = [
    'scatter',
    'pcolor_map',
    'pcolor_sphere',
    'balloon',
    'balloon_wireframe',
    'voronoi_cells_sphere',
    'contour',
    'contour_map',
    'MidpointNormalize',
    'phase_twilight',
]



""" Tests for sh channel indexing"""

import pytest
import numpy as np
import spharpy.indexing

def test_identity_matrix_n_nm():
    reference = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0],
                          [0, 1, 1, 1, 0, 0, 0, 0, 0],
                          [0, 0, 0, 0, 1, 1, 1, 1, 1]])

    identity = spharpy.indexing.sph_identity_matrix(2, type='n-nm')
    np.testing.assert_allclose(reference, identity)

def test_sid_indexing():
    n_max = 2
    reference_n = np.array([0, 1, 1, 1, 2, 2, 2, 2, 2])
    reference_m = np.array([0, 1, -1, 0, 2, -2, 1, -1, 0])


    sid_n, sid_m = spharpy.indexing.sid(n_max)

    np.testing.assert_equal(reference_n, sid_n)
    np.testing.assert_equal(reference_m, sid_m)

def test_sid2acn():
    n_max = 2
    # indexing starts at 0 here, reference was calculated
    # with indexing starting at 1.
    reference_acn = np.array([1, 3, 4, 2, 6, 8, 9, 7, 5]) - 1

    acn_indeces = spharpy.indexing.sid2acn(n_max)
    np.testing.assert_equal(reference_acn, acn_indeces)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The setup script.
"""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'numpy>=1.22',
    'scipy',
    'urllib3',
    'matplotlib>=3.3.0'
]

setup_requirements = [
    'pytest-runner',
]

test_requirements = [
    'pytest',
    'wheel',
    'flake8',
    'bump2version',
]


setup(
    name='spharpy',
    version='0.4.2',
    description="Python package for spherical array processing.",
    long_description=readme,
    author="Marco Berzborn",
    author_email='marco.berzborn@akustik.rwth-aachen.de',
    url='https://github.com/mberz/spharpy',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='spharpy',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    download_url="https://pypi.org/project/spharpy/",
    project_urls={
        "Bug Tracker": "https://github.com/mberz/spharpy/issues",
        "Documentation": "https://spharpy.readthedocs.io/",
        "Source Code": "https://github.com/mberz/spharpy",
    },
    python_requires='>=3.8'
)



import numpy as np
import numpy.polynomial as poly
from scipy.linalg import eig
from scipy.special import factorial

import spharpy
import spharpy.special as special


def dolph_chebyshev_weights(
        n_max,
        design_parameter,
        design_criterion='sidelobe'):
    """Calculate the weights for a spherical Dolph-Chebyshev beamformer. The
    design criterion can either be a desired side-lobe attenuation or a desired
    main-lobe width. Once one criterion is chosen, the other will become a
    dependent property which will be chosen accordingly.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    design_parameter : float, double
        This can either be the desired side-lobe attenuation or the width of
        the main-lobe in radians.
    design_criterion : 'sidelobe', 'mainlobe'
        Whether the design parameter argument is the desired side-lobe
        attenuation or the desired main-lobe width.

    Returns
    -------
    weigths : ndarray, double
        An array containing the weight coefficients $d_nm$.

    References
    ----------
    ..  [1] A. Koretz and B. Rafaely, “Dolph-Chebyshev beampattern design for
        spherical arrays,” IEEE Transactions on Signal Processing, vol. 57,
        no. 6, pp. 2417–2420, 2009.

    """
    M = 2*n_max
    if design_criterion == 'sidelobe':
        R = design_parameter
        x0 = np.cosh((1/M) * np.arccosh(R))
    elif design_criterion == 'mainlobe':
        theta0 = design_parameter
        x0 = np.cos(np.pi/2/M) / np.cos(theta0/2)
        R = np.cosh(M * np.arccosh(x0))
    else:
        raise ValueError("This design criterion is not available.")

    t_2N = special.chebyshev_coefficients(2*n_max)

    P_N = np.zeros((n_max+1, n_max+1))
    for n in range(n_max+1):
        P_N[0:n+1, n] = special.legendre_coefficients(n)

    d_n = np.zeros(n_max+1)
    for n in range(n_max+1):
        temp = 0
        for i in range(n+1):
            for j in range(n_max+1):
                for m in range(j+1):
                    temp = temp+(1-(-1)**(m+i+1))/(m+i+1) * \
                           factorial(j)/(factorial(m)*factorial(j-m)) * \
                           (1/2**j)*t_2N[2*j]*P_N[i, n]*x0**(2*j)
        d_n[n] = (2*np.pi/R)*temp

    weights = spharpy.indexing.sph_identity_matrix(n_max, type='n-nm').T @ d_n

    return weights


def rE_max_weights(n_max):
    """Weights that maximize the length of the energy vector.
    This is most often used in Ambisonics decoding.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order

    Returns
    -------
    weights : ndarray, double
        An array containing the weight coefficients.

    References
    ----------
    ..  [2] J. Daniel, J.-B. Rault, and J.-D. Polack, “Ambisonics Encoding of
        Other Audio Formats for Multiple Listening Conditions,” in 105th
        Convention of the Audio Engineering Society, 1998, vol. 3.

    """
    leg = poly.legendre.Legendre.basis(n_max+1)
    P_n_root = poly.legendre.legroots(leg.coef)
    max_root = np.max(np.abs(P_n_root))
    g_n = np.zeros(n_max+1)
    for n in range(0, n_max+1):
        leg = poly.legendre.Legendre.basis(n)
        g_n[n] = leg(max_root)

    weights = spharpy.indexing.sph_identity_matrix(n_max).T @ g_n

    return weights


def maximum_front_back_ratio_weights(n_max):
    """Weights that maximize the front-back ratio of the beam pattern.
    This is also often referred to as the super-cardioid beam pattern.

    Parameters
    ----------
    n_max : int
        The spherical harmonic order

    Returns
    -------
    weigths : ndarray, double
        An array containing the weight coefficients

    Note
    ----
    The weights are calculated from an eigenvalue problem

    References
    ----------
    [3] B. Rafaely, Fundamentals of Spherical Array Processing, Springer, 2015.

    """
    P_N = np.zeros((n_max+1, n_max+1))
    for n in range(n_max+1):
        P_N[0:n+1, n] = special.legendre_coefficients(n)

    Ann = np.zeros((n_max+1, n_max+1))
    Bnn = np.zeros((n_max+1, n_max+1))
    for n in range(n_max+1):
        for n_dash in range(n_max+1):
            const = 1/8/np.pi * (2*n+1) * (2*n_dash+1)
            temp = 0
            for q in range(0, n+1):
                for ll in range(0, n_dash+1):
                    temp += 1/(q+ll+1) * P_N[q, n] * P_N[ll, n_dash]
            Ann[n, n_dash] = temp * const

            temp = 0
            for q in range(0, n+1):
                for ll in range(0, n_dash+1):
                    temp += ((-1)**(q+ll))/(q+ll+1) * \
                        P_N[q, n] * P_N[ll, n_dash]
            Bnn[n, n_dash] = temp * const

    eigenvals, eigenvectors = eig(Ann, Bnn)
    f_n = eigenvectors[:, np.argmax(np.real(eigenvals))]
    weights = spharpy.indexing.sph_identity_matrix(n_max).T @ f_n

    return weights



"""
Helper functions for coordinate operations
"""

import numpy as np
from scipy.spatial import cKDTree, SphericalVoronoi


def sph2cart(r, theta, phi):
    """Transforms from spherical to Cartesian coordinates.
    Spherical coordinates follow the common convention in Physics/Mathematics
    Theta denotes the elevation angle with theta = 0 at the north pole and
    theta = pi at the south pole. Phi is the azimuth angle counting from
    phi = 0 at the x-axis in positive direction
    (counter clockwise rotation).

    .. math::

        x = r \\sin(\\theta) \\cos(\\phi),

        y = r \\sin(\\theta) \\sin(\\phi),

        z = r \\cos(\\theta)

    Parameters
    ----------
    r : ndarray, number
    theta : ndarray, number
    phi : ndarray, number

    Returns
    -------
    x : ndarray, number
    y : ndarray, number
    z : ndarray, number

    """
    x = r*np.sin(theta)*np.cos(phi)
    y = r*np.sin(theta)*np.sin(phi)
    z = r*np.cos(theta)
    return x, y, z


def cart2sph(x, y, z):
    """
    Transforms from Cartesian to spherical coordinates.
    Spherical coordinates follow the common convention in Physics/Mathematics
    Theta denotes the elevation angle with theta = 0 at the north pole and
    theta = pi at the south pole. Phi is the azimuth angle counting from
    phi = 0 at the x-axis in positive direction
    (counter clockwise rotation).

    .. math::

        r = \\sqrt{x^2 + y^2 + z^2},

        \\theta = \\arccos(\\frac{z}{r}),

        \\phi = \\arctan(\\frac{y}{x})

        0 < \\theta < \\pi,

        0 < \\phi < 2 \\pi


    Notes
    -----
    To ensure proper handling of the radiant for the azimuth angle, the arctan2
    implementatition from numpy is used here.

    Parameters
    ----------
    x : ndarray, number
    y : ndarray, number
    z : ndarray, number

    Returns
    -------
    r : ndarray, number
    theta : ndarray, number
    phi : ndarray, number

    """
    rad = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z/rad)
    phi = np.mod(np.arctan2(y, x), 2*np.pi)
    return rad, theta, phi


def cart2latlon(x, y, z):
    """Transforms from Cartesian coordinates to Geocentric coordinates

    .. math::

        h = \\sqrt{x^2 + y^2 + z^2},

        \\theta = \\pi/2 - \\arccos(\\frac{z}{r}),

        \\phi = \\arctan(\\frac{y}{x})

        -\\pi/2 < \\theta < \\pi/2,

        -\\pi < \\phi < \\pi

    where :math:`h` is the heigth, :math:`\\theta` is the latitude angle
    and :math:`\\phi` is the longitude angle

    Parameters
    ----------
    x : ndarray, number
        x-axis coordinates
    y : ndarray, number
        y-axis coordinates
    z : ndarray, number
        z-axis coordinates

    Returns
    -------
    height : ndarray, number
        The radius is rendered as height information
    latitude : ndarray, number
        Geocentric latitude angle
    longitude : ndarray, number
        Geocentric longitude angle

    """
    height = np.sqrt(x**2 + y**2 + z**2)
    latitude = np.pi/2 - np.arccos(z/height)
    longitude = np.arctan2(y, x)
    return height, latitude, longitude


def latlon2cart(height, latitude, longitude):
    """Transforms from Geocentric coordinates to Cartesian coordinates

    .. math::

        x = h \\cos(\\theta) \\cos(\\phi),

        y = h \\cos(\\theta) \\sin(\\phi),

        z = h \\sin(\\theta)

        -\\pi/2 < \\theta < \\pi/2,

        -\\pi < \\phi < \\pi

    where :math:`h` is the heigth, :math:`\\theta` is the latitude angle
    and :math:`\\phi` is the longitude angle

    Parameters
    ----------
    height : ndarray, number
        The radius is rendered as height information
    latitude : ndarray, number
        Geocentric latitude angle
    longitude : ndarray, number
        Geocentric longitude angle

    Returns
    -------
    x : ndarray, number
        x-axis coordinates
    y : ndarray, number
        y-axis coordinates
    z : ndarray, number
        z-axis coordinates

    """

    x = height * np.cos(latitude) * np.cos(longitude)
    y = height * np.cos(latitude) * np.sin(longitude)
    z = height * np.sin(latitude)

    return x, y, z


def spherical_voronoi(sampling, round_decimals=13, center=0.0):
    """Calculate a Voronoi diagram on the sphere for the given samplings
    points.

    Parameters
    ----------
    sampling : SamplingSphere
        Sampling points on a sphere
    round_decimals : int
        Number of decimals to be rounded to.
    center : double
        Center point of the voronoi diagram.

    Returns
    -------
    voronoi : SphericalVoronoi
        Spherical voronoi diagram as implemented in scipy.

    """
    points = sampling.cartesian.T
    radius = np.unique(np.round(sampling.radius, decimals=round_decimals))
    if len(radius) > 1:
        raise ValueError("All sampling points need to be on the \
                same radius.")
    voronoi = SphericalVoronoi(points, radius, center)

    return voronoi


def calculate_sampling_weights(sampling, round_decimals=12):
    """Calculate the sampling weights for numeric integration.

    Parameters
    ----------
    sampling : SamplingSphere
        Sampling points on a sphere
    round_decimals : int, optional

    apply : boolean, optional
        Whether or not to store the weights into the class object

    Returns
    -------
    weigths : ndarray, float
        Sampling weights

    """
    sv = spherical_voronoi(sampling, round_decimals=round_decimals)
    sv.sort_vertices_of_regions()

    unique_verts, idx_uni = np.unique(
        np.round(sv.vertices, decimals=10),
        axis=0,
        return_index=True)

    searchtree = cKDTree(unique_verts)
    area = np.zeros(sampling.n_points, float)

    for idx, region in enumerate(sv.regions):
        _, idx_nearest = searchtree.query(sv.vertices[np.array(region)])
        mask_unique = np.sort(np.unique(idx_nearest, return_index=True)[1])
        mask_new = idx_uni[idx_nearest[mask_unique]]

        area[idx] = _poly_area(sv.vertices[mask_new])

    area = area / np.sum(area) * 4 * np.pi

    return area


def _unit_normal(a, b, c):
    x = np.linalg.det(
        [[1, a[1], a[2]],
         [1, b[1], b[2]],
         [1, c[1], c[2]]])
    y = np.linalg.det(
        [[a[0], 1, a[2]],
         [b[0], 1, b[2]],
         [c[0], 1, c[2]]])
    z = np.linalg.det(
        [[a[0], a[1], 1],
         [b[0], b[1], 1],
         [c[0], c[1], 1]])

    magnitude = np.sqrt(x**2 + y**2 + z**2)

    return (x/magnitude, y/magnitude, z/magnitude)


def _poly_area(poly):
    # area of polygon poly
    if len(poly) < 3:
        # not a plane - no area
        return 0
    total = [0.0, 0.0, 0.0]
    N = len(poly)
    for i in range(N):
        vi1 = poly[i]
        vi2 = poly[np.mod((i+1), N)]
        prod = np.cross(vi1, vi2)
        total[0] += prod[0]
        total[1] += prod[1]
        total[2] += prod[2]
    result = np.dot(total, _unit_normal(poly[0], poly[1], poly[2]))
    return np.abs(result/2)



""" Tests for sh rotations"""

import numpy as np
import spharpy.transforms as transforms
from spharpy.spherical import spherical_harmonic_basis
from spharpy.samplings import Coordinates
import spharpy
from spharpy.transforms import RotationSH
import pytest


def test_rotation_matrix_z_axis_complex():
    rot_angle = np.pi/2
    n_max = 2
    reference = np.diag([1, 1j, 1, -1j, -1, 1j, 1, -1j, -1])

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    np.testing.assert_allclose(rot_mat, reference)

    rot_angle = np.pi
    reference = np.diag([1, -1, 1, -1, 1, -1, 1, -1, 1])

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    np.testing.assert_allclose(rot_mat, reference)

    rot_angle = 3/2*np.pi
    reference = np.diag([1, -1j, 1, 1j, -1, -1j, 1, 1j, -1])

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    np.testing.assert_allclose(rot_mat, reference)


def test_rotation_sh_basis_z_axis_complex():
    rot_angle = np.pi/2
    n_max = 2
    coords_x = Coordinates(1, 0, 0)
    coords_y = Coordinates(0, 1, 0)
    reference = spherical_harmonic_basis(n_max, coords_y).T.conj()

    rot_mat = transforms.rotation_z_axis(n_max, rot_angle)
    sh_vec_x = spherical_harmonic_basis(n_max, coords_x)
    sh_vec_rotated = rot_mat @ sh_vec_x.T.conj()

    sampling = spharpy.samplings.dodecahedron()
    Y = spharpy.spherical.spherical_harmonic_basis(n_max, sampling)

    np.testing.assert_allclose(Y@sh_vec_rotated, Y@reference)


def test_rotation_maxtrix_z_axis_real():
    n_max = 2
    rot_angle = np.pi/4
    reference = np.array([
        [1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, np.cos(rot_angle), 0, np.sin(rot_angle), 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -np.sin(rot_angle), 0, np.cos(rot_angle), 0, 0, 0, 0, 0],
        [0, 0, 0, 0, np.cos(2*rot_angle), 0, 0, 0, np.sin(2*rot_angle)],
        [0, 0, 0, 0, 0, np.cos(rot_angle), 0, np.sin(rot_angle), 0],
        [0, 0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, -np.sin(rot_angle), 0, np.cos(rot_angle), 0],
        [0, 0, 0, 0, -np.sin(2*rot_angle), 0, 0, 0, np.cos(2*rot_angle)]])
    rot_matrix = transforms.rotation_z_axis_real(n_max, rot_angle)
    np.testing.assert_allclose(rot_matrix, reference)


def test_rotation_sh_basis_z_axis_real():
    rot_angle = np.pi/2
    n_max = 2
    coords_x = Coordinates(1, 0, 0)
    coords_y = Coordinates(0, 1, 0)
    reference = np.squeeze(
        spharpy.spherical.spherical_harmonic_basis_real(n_max, coords_y))

    rot_mat = spharpy.transforms.rotation_z_axis_real(n_max, rot_angle)
    sh_vec_x = spharpy.spherical.spherical_harmonic_basis_real(n_max, coords_x)
    sh_vec_rotated = np.squeeze(rot_mat @ np.squeeze(sh_vec_x))

    np.testing.assert_almost_equal(sh_vec_rotated, np.squeeze(reference))


def test_wigner_d_rot_real():
    euler_angles = []
    n_max = 5
    euler_angles = np.deg2rad(np.array([90., 45., -90.]))
    D = spharpy.transforms.wigner_d_rotation_real(
        n_max, euler_angles[0], euler_angles[1], euler_angles[2])

    reference = np.squeeze(spharpy.spherical.spherical_harmonic_basis_real(
            n_max, Coordinates(0, 1, 1)))

    sh_vec = np.squeeze(spharpy.spherical.spherical_harmonic_basis_real(
        n_max, Coordinates(0, 1, 0)))
    sh_vec_rotated = D @ sh_vec

    np.testing.assert_allclose(sh_vec_rotated, reference, atol=1e-10)

    rot_angle_z = np.pi/4
    D = spharpy.transforms.wigner_d_rotation_real(
        n_max, rot_angle_z, 0, 0)

    rot_mat = spharpy.transforms.rotation_z_axis_real(n_max, rot_angle_z)

    np.testing.assert_allclose(D, rot_mat, atol=1e-7)

    pass


def test_wigner_d_rot():
    euler_angles = []
    n_max = 5
    euler_angles = np.deg2rad(np.array([90., 45., -90.]))
    D = spharpy.transforms.wigner_d_rotation(
        n_max, euler_angles[0], euler_angles[1], euler_angles[2])

    reference = np.squeeze(spharpy.spherical.spherical_harmonic_basis(
            n_max, Coordinates(0, 1, 1)))

    sh_vec = np.squeeze(spharpy.spherical.spherical_harmonic_basis(
        n_max, Coordinates(0, 1, 0)))
    sh_vec_rotated = D @ sh_vec

    np.testing.assert_allclose(sh_vec_rotated, reference)

    rot_angle_z = np.pi/4
    D = spharpy.transforms.wigner_d_rotation(
        n_max, rot_angle_z, 0, 0)

    rot_mat = spharpy.transforms.rotation_z_axis(n_max, rot_angle_z)

    np.testing.assert_allclose(D, rot_mat, atol=1e-7)


def test_RotationSH():
    n_max = 4
    rot_angle_z = np.pi/2
    rot_vec = [0, 0, rot_angle_z]
    rot = RotationSH.from_rotvec(n_max, rot_vec)

    assert rot._n_max == n_max

    n_max = 2
    rot.n_max = n_max
    assert rot._n_max == n_max
    assert rot.n_max == n_max

    with pytest.raises(ValueError):
        rot.n_max = -1

    D_Rot = rot.as_spherical_harmonic(type='real')

    reference = np.array([
        [1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, np.cos(rot_angle_z), 0, np.sin(rot_angle_z), 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -np.sin(rot_angle_z), 0, np.cos(rot_angle_z), 0, 0, 0, 0, 0],
        [0, 0, 0, 0, np.cos(2*rot_angle_z), 0, 0, 0, np.sin(2*rot_angle_z)],
        [0, 0, 0, 0, 0, np.cos(rot_angle_z), 0, np.sin(rot_angle_z), 0],
        [0, 0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, -np.sin(rot_angle_z), 0, np.cos(rot_angle_z), 0],
        [0, 0, 0, 0, -np.sin(2*rot_angle_z), 0, 0, 0, np.cos(2*rot_angle_z)]])

    np.testing.assert_allclose(D_Rot, reference, atol=1e-10)

    rot = RotationSH.from_rotvec(n_max, [0, 0, 90], degrees=True)
    np.testing.assert_allclose(
        rot.as_spherical_harmonic(type='real'),
        reference, atol=1e-10)

    rot = RotationSH.from_euler(n_max, 'zyz', [0, 0, 90], degrees=True)
    np.testing.assert_allclose(
        rot.as_spherical_harmonic(type='real'),
        reference, atol=1e-10)

    rot = RotationSH.from_quat(n_max, [0, 0, 1/np.sqrt(2), 1/np.sqrt(2)])
    np.testing.assert_allclose(
        rot.as_spherical_harmonic(type='real'),
        reference, atol=1e-10)

    rot_mat_z_spat = np.array([
        [0, -1, 0],
        [1, 0, 0],
        [0, 0, 1]])
    rot = RotationSH.from_matrix(n_max, rot_mat_z_spat)
    np.testing.assert_allclose(
        rot.as_spherical_harmonic(type='real'),
        reference, atol=1e-7)

    # mrp = [0, 0, np.tan(np.pi/2/4)]
    # rot = RotationSH.from_mrp(n_max, mrp)
    # np.testing.assert_allclose(
    #     rot.as_spherical_harmonic(type='real'),
    #     reference, atol=1e-7)
